
def ls():
    pass


def load(Loader, representation):
    pass


def create(name, family, asset, options, data):
    pass


def update(container, version):
    pass


def remove(container):
    pass
