from .pipeline import (
    install,
    uninstall,

    Creator,

    ls,
    containerise,

)

from .workio import (
    open_file,
    save_file,
    current_file,
    has_unsaved_changes,
    file_extensions,
    work_root
)

from .lib import (
    lsattr,
    lsattrs,
    read,

    maintained_selection,
    unique_name
)


__all__ = [
    "install",
    "uninstall",

    "Creator",

    "ls",
    "containerise",

    # Workfiles API
    "open_file",
    "save_file",
    "current_file",
    "has_unsaved_changes",
    "file_extensions",
    "work_root",

    # Utility functions
    "lsattr",
    "lsattrs",
    "read",

    "maintained_selection",
    "unique_name"
]

# Backwards API compatibility
open = open_file
save = save_file
